"""
An opionated set of helpers for defining Docker integration test environments with py.test fixtures.
"""

from .factories import build, container, fetch, image, image_or_build, network, volume
from .utils import fxtr

__version__ = "0.2.3"


__all__ = [
    "build",
    "container",
    "fetch",
    "image",
    "image_or_build",
    "network",
    "volume",
    "fxtr",
]



from .build import build
from .container import container
from .fetch import fetch
from .image import image
from .image_or_build import image_or_build
from .network import network
from .volume import volume

__all__ = [
    "build",
    "container",
    "fetch",
    "image",
    "image_or_build",
    "network",
    "volume",
]



"""
This module contains a wrapper that adds some helpers to a Docker Container
object that are useful for integration testing.
"""

import io
import tarfile

from pytest_docker_tools.exceptions import (
    ContainerFailed,
    ContainerNotReady,
    TimeoutError,
)
from pytest_docker_tools.utils import tests_inside_container, wait_for_callable


class _Map:
    def __init__(self, container):
        self._container = container

    def values(self):
        return [self[k] for k in self.keys()]

    def items(self):
        return [(k, self[k]) for k in self.keys()]

    def __iter__(self):
        return iter(self.keys())


class IpMap(_Map):
    @property
    def primary(self):
        return next(iter(self.values()))

    def keys(self):
        return self._container.attrs["NetworkSettings"]["Networks"].keys()

    def __getitem__(self, key):
        if not isinstance(key, str):
            key = key.name

        networks = self._container.attrs["NetworkSettings"]["Networks"]
        if key not in networks:
            raise KeyError(f"Unknown network: {key}")

        return networks[key]["IPAddress"]


class PortMap(_Map):
    def __init__(self, container):
        self._container = container

    def keys(self):
        return self._container.attrs["NetworkSettings"]["Ports"].keys()

    def __getitem__(self, key):
        ports = self._container.attrs["NetworkSettings"]["Ports"]
        if key not in ports:
            raise KeyError(f"Unknown port: {key}")

        if not ports[key]:
            return []

        return [int(p["HostPort"]) for p in ports[key]]


class Container:
    def __init__(self, container):
        self._container = container
        self.ips = IpMap(container)
        self.ports = PortMap(container)

    def ready(self):
        self._container.reload()

        if self.status == "exited":
            raise ContainerFailed(
                self,
                f"Container {self.name} has already exited before we noticed it was ready",
            )

        if self.status != "running":
            return False

        networks = self._container.attrs["NetworkSettings"]["Networks"]
        for name, network in networks.items():
            if not network["IPAddress"]:
                return False

        # If a user has exposed a port then wait for LISTEN socket to show up in netstat
        ports = self._container.attrs["NetworkSettings"]["Ports"]
        for port, listeners in ports.items():
            if not listeners:
                if port in (self._container.attrs["HostConfig"]["PortBindings"] or []):
                    return False
                continue

            port, proto = port.split("/")

            assert proto in ("tcp", "udp")

            if proto == "tcp" and port not in self.get_open_tcp_ports():
                return False

            if proto == "udp" and port not in self.get_open_udp_ports():
                return False

        return True

    @property
    def attrs(self):
        return self._container.attrs

    @property
    def id(self):
        return self._container.id

    @property
    def name(self):
        return self._container.name

    @property
    def env(self):
        kv_pairs = map(
            lambda v: v.split("=", 1), self._container.attrs["Config"]["Env"]
        )
        return {k: v for k, v in kv_pairs}

    @property
    def status(self):
        return self._container.status

    def exec_run(self, *args, **kwargs):
        return self._container.exec_run(*args, **kwargs)

    def reload(self):
        return self._container.reload()

    def restart(self, timeout=10):
        self._container.restart(timeout=timeout)

        try:
            wait_for_callable(
                "Waiting for container to be ready after restart", self.ready
            )
        except TimeoutError:
            raise ContainerNotReady(
                self, "Timeout while waiting for container to be ready after restart"
            )

    def kill(self, signal=None):
        return self._container.kill(signal)

    def remove(self, *args, **kwargs):
        raise RuntimeError(
            "Do not remove this container manually. It will be removed automatically by py.test after the test finishes."
        )

    def logs(self):
        return self._container.logs().decode("utf-8")

    def get_files(self, path):
        """
        Retrieve files from a container at a given path.

        This is meant for extracting log files from a container where it is not
        using the docker logging capabilities.
        """

        archive_iter, _ = self._container.get_archive(path)

        archive_stream = io.BytesIO()
        [archive_stream.write(chunk) for chunk in archive_iter]
        archive_stream.seek(0)

        archive = tarfile.TarFile(fileobj=archive_stream)
        files = {}
        for info in archive.getmembers():
            if not info.isfile():
                files[info.name] = None
                continue
            reader = archive.extractfile(info.name)
            files[info.name] = reader.read()

        return files

    def get_text(self, path):
        text = {}
        for path, bytes in self.get_files(path).items():
            if bytes is None:
                text[path] = None
                continue
            text[path] = bytes.decode("utf-8")
        return text

    def get_open_tcp_ports(self):
        """Gets all TCP sockets in the LISTEN state"""
        netstat = (
            self._container.exec_run("cat /proc/net/tcp /proc/net/tcp6")[1]
            .decode("utf-8")
            .strip()
        )

        ports = []
        for line in netstat.split("\n"):
            # Not interested in empty lines
            if not line:
                continue

            line = line.split()

            # Only interested in listen sockets
            if line[3] != "0A":
                continue

            ports.append(str(int(line[1].split(":", 1)[1], 16)))

        return ports

    def get_open_udp_ports(self):
        """Gets all UDP sockets in the LISTEN state"""
        netstat = (
            self._container.exec_run("cat /proc/net/udp /proc/net/udp6")[1]
            .decode("utf-8")
            .strip()
        )

        ports = []
        for line in netstat.split("\n"):
            # Not interested in empty lines
            if not line:
                continue

            line = line.split()

            # If we are listening on a UDP port it will appear in /proc/net/udp
            # and state will be '07'
            if line[3] != "07":
                continue

            ports.append(str(int(line[1].split(":", 1)[1], 16)))

        return ports

    def get_addr(self, port):
        if tests_inside_container():
            return (self.ips.primary, int(port.split("/")[0]))
        else:
            return ("127.0.0.1", self.ports[port][0])



import base64
import hashlib
import json
import os
import sys
import time
from typing import Any, Callable, Dict

from _pytest.fixtures import SubRequest
from docker.models.containers import Container
from docker.models.networks import Network
from docker.models.volumes import Volume

from .exceptions import TimeoutError

LABEL_REUSABLE = "pytest-docker-tools.reusable"
LABEL_SIGNATURE = "pytest-docker-tools.signature"


def wait_for_callable(message: str, func: Callable, timeout: int = 30) -> None:
    """
    Runs a callable once a second until it returns True or we hit the timeout.
    """
    sys.stdout.write(message)
    try:
        for i in range(timeout):
            sys.stdout.write(".")
            sys.stdout.flush()

            if func():
                return

            time.sleep(1)
    finally:
        sys.stdout.write("\n")

    raise TimeoutError(f"Timeout of {timeout}s exceeded")


def tests_inside_container() -> bool:
    """Returns True if tests are running inside a Linux container"""

    return os.path.isfile("/.dockerenv") or os.path.isfile("/run/.containerenv")


def is_reusable_container(container: Container) -> bool:
    return container.attrs["Config"]["Labels"].get(LABEL_REUSABLE) == "True"


def is_reusable_network(network: Network) -> bool:
    return network.attrs["Labels"].get(LABEL_REUSABLE) == "True"


def is_reusable_volume(volume: Volume) -> bool:
    labels = volume.attrs["Labels"]
    return labels and labels.get(LABEL_REUSABLE) == "True"


def set_reusable_labels(kwargs: Dict[str, Any], request: SubRequest) -> None:
    labels = kwargs.setdefault("labels", {})

    labels.update(
        {
            "creator": "pytest-docker-tools",
            LABEL_REUSABLE: str(request.config.option.reuse_containers),
        }
    )


class Base64Encoder(json.JSONEncoder):
    def default(self, obj: Any) -> str:
        if isinstance(obj, bytes):
            return base64.b64encode(obj).decode("utf-8")
        return super().default(obj)


def hash_params(kwargs: Dict[str, Any]) -> str:
    rendered = json.dumps(kwargs, cls=Base64Encoder, sort_keys=True).encode("utf-8")
    signature = hashlib.sha256(rendered).hexdigest()
    return signature


def set_signature(kwargs: Dict[str, Any], signature: str) -> None:
    labels = kwargs.setdefault("labels", {})
    labels[LABEL_SIGNATURE] = signature


def check_signature(labels: Dict[str, Any], signature: str) -> bool:
    return labels.get(LABEL_SIGNATURE, "") == signature


def is_using_network(container: Container, network: Network) -> bool:
    settings = container.attrs.get("NetworkSettings", {})
    return network.name in settings.get("Networks", {})


def is_using_volume(container: Container, volume: Volume) -> bool:
    for mount in container.attrs.get("Mounts", []):
        if mount["Type"] != "volume":
            continue
        if mount["Name"] == volume.name:
            return True
    return False


class _FixtureRef:
    def __init__(self, name: str) -> None:
        self._name = name

    @property
    def name(self) -> str:
        return self._name


fxtr = _FixtureRef



import sys

from pytest_docker_tools.builder import fixture_factory


@fixture_factory(scope="session")
def build(request, docker_client, wrapper_class, **kwargs):
    """Docker image: built from "{path}" """

    # The docker build command now defaults to --rm=true, but docker-py doesn't
    # Let's do what docker build does by default
    kwargs.setdefault("rm", True)

    if "path" in kwargs:
        sys.stdout.write(f"Building {kwargs['path']}")
    else:
        sys.stdout.write("Building")

    # If specified, build these stages and tag them first
    # The main image build works without doing this but this allows us to preserve
    # 'builder' stages when doing 'docker image prune'
    stages = kwargs.pop("stages", {})
    for stage, tag in stages.items():
        stage_kwargs = kwargs.copy()
        stage_kwargs["tag"] = tag
        stage_kwargs["target"] = stage

        image, logs = docker_client.images.build(**stage_kwargs)

        for line in logs:
            sys.stdout.write(".")
            sys.stdout.flush()

    # Build the image
    try:
        image, logs = docker_client.images.build(**kwargs)

        for line in logs:
            sys.stdout.write(".")
            sys.stdout.flush()

    finally:
        sys.stdout.write("\n")

    # request.addfinalizer(lambda: docker_client.images.remove(image.id))

    wrapper_class = wrapper_class or (lambda image: image)
    return wrapper_class(image)



import uuid

from docker.client import DockerClient
from docker.errors import NotFound
import pytest

from pytest_docker_tools.builder import fixture_factory
from pytest_docker_tools.utils import (
    check_signature,
    hash_params,
    is_reusable_container,
    is_reusable_network,
    is_using_network,
    set_reusable_labels,
    set_signature,
)


def _remove_stale_network(network):
    for container in network.client.containers.list(ignore_removed=True, all=True):
        if not is_using_network(container, network):
            continue

        if not is_reusable_container(container):
            pytest.fail(
                f"The network {network.name} is connected to a non-reusable container: {container.id}"
            )

        print(
            f"Removing container {container.name} connected to stale network {network.name}"
        )
        container.remove(force=True)

    print(f"Removing stale reusable network: {network.name}")
    network.remove()


@fixture_factory()
def network(request, docker_client: DockerClient, wrapper_class, **kwargs):
    """Docker network"""

    set_reusable_labels(kwargs, request)

    signature = hash_params(kwargs)
    set_signature(kwargs, signature)

    wrapper_class = wrapper_class or (lambda network: network)

    if request.config.option.reuse_containers:
        if "name" not in kwargs.keys():
            pytest.fail(
                "Error: Tried to use '--reuse-containers' command line argument without "
                "setting 'name' attribute on network"
            )

        name = kwargs["name"]
        try:
            network = docker_client.networks.get(name)
        except NotFound:
            pass
        else:
            # Found a network with the right name, but it doesn't have pytest-docker-tools labels
            # We shouldn't just clobber it, its not ours. Bail out.
            if not is_reusable_network(network):
                pytest.fail(
                    f"Tried to reuse {network.name} but it does not appear to be a reusable network"
                )

            # It's ours, and its not stale. Reuse it!
            if check_signature(network.attrs["Labels"], signature):
                return wrapper_class(network)

            # It's ours and it is stale. Clobber it.
            _remove_stale_network(network)

    name = kwargs.pop("name", "pytest-{uuid}").format(uuid=str(uuid.uuid4()))

    print(f"Creating network {name}")
    network = docker_client.networks.create(name, **kwargs)

    if not request.config.option.reuse_containers:
        request.addfinalizer(lambda: network.remove())

    return wrapper_class(network)



from typing import Any

from docker.models.containers import Container


class TimeoutError(Exception):
    pass


class ContainerError(Exception):
    def __init__(self, container: Container, *args: Any, **kwargs: Any) -> None:
        self._container = container
        super().__init__(*args, **kwargs)


class ContainerFailed(ContainerError):
    pass


class ContainerNotReady(ContainerError):
    pass



import io
import os
import tarfile
import uuid

from docker.errors import NotFound
import pytest

from pytest_docker_tools.builder import fixture_factory
from pytest_docker_tools.utils import (
    check_signature,
    hash_params,
    is_reusable_container,
    is_reusable_volume,
    is_using_volume,
    set_reusable_labels,
    set_signature,
)


def _remove_stale_volume(docker_client, volume):
    for container in docker_client.containers.list(ignore_removed=True, all=True):
        if not is_using_volume(container, volume):
            continue

        if not is_reusable_container(container):
            pytest.fail(
                f"The volume {volume.name} is connected to a non-reusable container: {container.id}"
            )

        print(
            f"Removing container {container.name} connected to stale volume {volume.name}"
        )
        container.remove(force=True)

    print(f"Removing stale reusable volume: {volume.name}")
    volume.remove()


def _populate_volume(docker_client, volume, seeds):
    fp = io.BytesIO()
    tf = tarfile.open(mode="w:gz", fileobj=fp)

    for path, contents in seeds.items():
        ti = tarfile.TarInfo(path)
        if contents is None:
            ti.type = tarfile.DIRTYPE
            tf.addfile(ti)
        else:
            ti.size = len(contents)
            tf.addfile(ti, io.BytesIO(contents))

    tf.close()
    fp.seek(0)

    image, logs = docker_client.images.build(
        path=os.path.join(os.path.dirname(__file__), "..", "contexts/scratch"),
        rm=True,
    )
    list(logs)

    container = docker_client.containers.create(
        image=image.id,
        volumes={
            f"{volume.name}": {"bind": "/data"},
        },
    )

    try:
        container.put_archive("/data", fp)
    finally:
        container.remove(force=True)


@fixture_factory()
def volume(request, docker_client, wrapper_class, **kwargs):
    """Docker volume"""
    wrapper_class = wrapper_class or (lambda volume: volume)

    set_reusable_labels(kwargs, request)

    signature = hash_params(kwargs)
    set_signature(kwargs, signature)

    if request.config.option.reuse_containers:
        if "name" not in kwargs.keys():
            pytest.fail(
                "Error: Tried to use '--reuse-containers' command line argument without "
                "setting 'name' attribute on volume"
            )

        name = kwargs["name"]
        try:
            volume = docker_client.volumes.get(name)
        except NotFound:
            pass
        else:
            # Found a volume with the right name, but it doesn't have pytest-docker-tools labels
            # We shouldn't just clobber it, its not ours. Bail out.
            if not is_reusable_volume(volume):
                pytest.fail(
                    f"Tried to reuse {name} but it does not appear to be a reusable volume"
                )

            # It's ours, and its not stale. Reuse it!
            if check_signature(volume.attrs["Labels"], signature):
                return wrapper_class(volume)

            # It's ours and it is stale. Clobber it.
            _remove_stale_volume(docker_client, volume)

    name = kwargs.pop("name", "pytest-{uuid}").format(uuid=str(uuid.uuid4()))
    seeds = kwargs.pop("initial_content", {})

    print(f"Creating volume {name}")
    volume = docker_client.volumes.create(name, **kwargs)

    if not request.config.option.reuse_containers:
        request.addfinalizer(lambda: volume.remove(True))

    if seeds:
        _populate_volume(docker_client, volume, seeds)

    return wrapper_class(volume)



import os

from .build import build
from .image import image


def image_or_build(environ_key, **kwargs):
    if environ_key in os.environ:
        return image(name=os.environ[environ_key])
    return build(**kwargs)



import docker
import pytest

from .exceptions import ContainerError
from .wrappers import Container


@pytest.fixture(scope="session")
def docker_client(request) -> docker.client.DockerClient:
    """A Docker client configured from environment variables"""
    return docker.from_env()


@pytest.hookimpl(tryfirst=True, hookwrapper=True)
def pytest_runtest_makereport(item, call):
    """
    This hook allows Docker containers to contribute their logs to the py.test
    report.
    """

    outcome = yield
    rep = outcome.get_result()

    if not rep.failed:
        return

    if call.excinfo and isinstance(call.excinfo.value, ContainerError):
        container = call.excinfo.value._container
        rep.sections.append((container.name, container.logs()))

    if "request" not in item.funcargs:
        return

    for name, fixturedef in item.funcargs["request"]._fixture_defs.items():
        if not hasattr(fixturedef, "cached_result") or not fixturedef.cached_result:
            continue
        fixture = fixturedef.cached_result[0]
        if isinstance(fixture, Container):
            rep.sections.append(
                (
                    name + ": " + fixture.name,
                    fixture.logs(),
                )
            )


def pytest_addoption(parser):
    group = parser.getgroup("pytest-docker-tools", "Pytest Docker Tools")
    group.addoption(
        "--reuse-containers",
        action="store_true",
        dest="reuse_containers",
        help="reuse existing containers instead of always creating new ones. Requires the 'name' attribute to be set"
        "on container definition",
    )



from pytest_docker_tools.builder import fixture_factory


@fixture_factory(scope="session")
def image(request, docker_client, wrapper_class, **kwargs):
    """Docker image: named "{name}" (already available)"""

    image = docker_client.images.get(kwargs["name"])
    wrapper_class = wrapper_class or (lambda image: image)
    return wrapper_class(image)



from docker.errors import NotFound
import pytest

from pytest_docker_tools.builder import fixture_factory
from pytest_docker_tools.exceptions import ContainerNotReady, TimeoutError
from pytest_docker_tools.utils import (
    check_signature,
    hash_params,
    is_reusable_container,
    set_reusable_labels,
    set_signature,
    wait_for_callable,
)
from pytest_docker_tools.wrappers import Container


@fixture_factory()
def container(request, docker_client, wrapper_class, **kwargs):
    """Docker container: image={image}"""

    wrapper_class = wrapper_class or Container

    kwargs.update({"detach": True})
    set_reusable_labels(kwargs, request)

    signature = hash_params(kwargs)
    set_signature(kwargs, signature)

    if request.config.option.reuse_containers:
        if "name" not in kwargs.keys():
            pytest.fail(
                "Tried to use '--reuse-containers' command line argument without "
                "setting 'name' attribute on container"
            )

        name = kwargs["name"]

        try:
            current = docker_client.containers.get(name)
        except NotFound:
            pass
        else:
            # Found a container with the right name, but it doesn't have pytest-docker-tools labels
            # We shouldn't just clobber it, its not ours. Bail out.
            if not is_reusable_container(current):
                pytest.fail(
                    f"Tried to reuse {name} but it does not appear to be a reusable container"
                )

            # It's ours, and its not stale. Reuse it!
            if check_signature(current.labels, signature):
                return wrapper_class(current)

            # It's ours and it is stale. Clobber it.
            print(f"Removing stale reusable container: {name}")
            current.remove(force=True)

    timeout = kwargs.pop("timeout", 30)

    raw_container = docker_client.containers.run(**kwargs)
    if not request.config.option.reuse_containers:
        request.addfinalizer(
            lambda: raw_container.remove(force=True) and raw_container.wait(timeout=10)
        )

    container = wrapper_class(raw_container)

    try:
        wait_for_callable("Waiting for container to be ready", container.ready, timeout)
    except TimeoutError:
        raise ContainerNotReady(
            container, "Timeout while waiting for container to be ready"
        )

    return container



import textwrap

from .templates import find_fixtures_in_params, resolve_fixtures_in_params


def build_fixture_function(callable, scope, wrapper_class, kwargs):
    name = callable.__name__
    docstring = [
        "An instance of {}.{} created with the following:\n".format(
            name, callable.__module__
        )
    ]
    docstring.extend(f"        {key}: {value}" for key, value in kwargs.items())
    docstring = "\n".join(docstring)

    fixtures = find_fixtures_in_params(kwargs).union({"request", "docker_client"})
    fixtures_str = ",".join(fixtures)

    template = textwrap.dedent(
        f"""
    import pytest

    def {name}({fixtures_str}):
        r\'\'\'
        {docstring}
        \'\'\'
        real_kwargs = resolve_fixtures_in_params(request, kwargs)
        return _{name}(request, docker_client, wrapper_class=wrapper_class, **real_kwargs)

    {name}.__module__ = callable.__module__
    {name} = pytest.fixture(scope=scope)({name})
    """
    )
    globals = {
        "resolve_fixtures_in_params": resolve_fixtures_in_params,
        f"_{name}": callable,
        "kwargs": kwargs,
        "scope": scope,
        "wrapper_class": wrapper_class,
    }
    exec(template, globals)
    return globals[name]


def fixture_factory(scope="function"):
    wrapper_class = None

    def inner(callable):
        def factory(*, scope=scope, wrapper_class=wrapper_class, **kwargs):
            return build_fixture_function(callable, scope, wrapper_class, kwargs)

        factory.__name__ = callable.__name__
        factory.__doc__ = getattr(callable, "__doc__", "")
        return factory

    return inner
